import numpy, random, pygame, bezier
from operator import add

class TileType (object):
	def __init__(self, tid, name, button, natural, color, base_cost, income, buildable, options={}, neighbors = None):
		self.tid = tid
		self.name = name
		self.button = button
		self.natural = natural
		self.color = color
		self.base_cost = base_cost
		self.income = income
		self.buildable = buildable
		self.neighbors = (neighbors is None) and [name] or neighbors
		self.options = options

class TerrainParameters (object):
	"""Just placeholder for named terrain parameters.
	
	All parameters are specified as a tuple with min and max values.
	"""

	forests_number = (20, 30)
	forests_radius = (4, 10)
	river_width = (2, 3)

class TerrainInfo (object):
	"""All terrain metainformation: terrain types, generation parameters etc."""
	def __init__(self):
		self.next_id = 0
		self.types = {}
		self.names = {}
		self.params = None

	def add_type(self, *params):
		t = TileType(self.next_id, *params)
		self.next_id += 1
		self.types[t.tid] = t
		if t.name in self.names:
			raise KeyError, "duplacate terrian name"
		self.names[t.name] = t

	def get_type(self, name):
		return self.names[name]


class MildTerrain (TerrainInfo):
	land_price = 50.0
	city_radius = 20
	city_modifier_base = 1.5

	def __init__(self):
		TerrainInfo.__init__(self)

		self.add_type("grass", None, None, (64, 192, 0), 5.0, 0.0, False, {}, [])
		self.add_type("forest", None, None, (0, 128, 0), 100.0, 0.0, False, {}, ['forest', 'park'])
		self.add_type("water", None, None, (0, 0, 128), 1000.0, 0.0, False)
		self.add_type("city", None, None, (64, 0, 0), 0.0, 0.0, False)

		self.add_type("field", "Start a farm", "grass",
				(64, 64, 0), 50.0, 50.0, True)
		self.add_type("golf", "Build a golf field", "grass",
				(128, 224, 64), 200.0, 100., True, { 15: 14 })
		self.add_type("park", "Build a park", "forest",
				(128, 192, 128), 100.0, -50.0, True, {}, ['forest', 'park'])

		self.params = TerrainParameters

class Direction(object):
	def __init__(self, dx, dy, flag, other, symbol):
		self.dx = dx
		self.dy = dy
		self.flag = flag
		self.other = other
		self.symbol = symbol

	def __repr__(self):
		return '<Direction %s>' % self.symbol

	def calc(self, x, y, length = 1):
		return (x + self.dx * length, y + self.dy * length)


DIR_ALL = 15
DIR_N = Direction( 0, -1, 1, 4, 'N')
DIR_E = Direction( 1,  0, 2, 8, 'E')
DIR_S = Direction( 0,  1, 4, 1, 'S')
DIR_W = Direction(-1,  0, 8, 2, 'W')
DIRECTIONS = [ DIR_N, DIR_E, DIR_S, DIR_W ]

OWNER_MASK = 3

class Map (object):
	"""Stores the world and related information.
	"""
	def __init__(self, w, h, terraininfo):
		self.w = w
		self.h = h
		self.terraininfo = terraininfo
		self.types = terraininfo.types
		
		self.terrain = numpy.zeros((w, h), dtype=numpy.uint8)
		self.flags = numpy.zeros((w, h), dtype=numpy.uint8)
		self.cities = numpy.zeros((w, h), dtype=numpy.uint8)
		self.sprites = numpy.zeros((w, h), dtype=numpy.uint8)

		self.neigh_map = {}
		for tid, type in self.types.items():
			self.neigh_map[tid] = [terraininfo.get_type(n).tid for n in type.neighbors]

		self.dirty = []
		self.generate_terrain()
		self.dirty = []
		
	
	def generate_terrain(self):
		"""Generates terrain.

		params should be either TerrainParameters class or anything with
		the same attributes.

		Does not reset the state.
		Returns nothing.

		(This actually belongs to TerrainInfo classes, but since there's
		only one terrain type, it doesn't matter a lot.)
		"""
		params = self.terraininfo.params

		# Place forests in chunks.
		for n in range(random.randint(*params.forests_number)):
			self.place_random_forest(params)

		self.create_river(params)

	def draw_terrain_circle(self, cx, cy, radius, terrain):
		radius2 = radius ** 2
		for x in range(cx - radius - 1, cx + radius + 1):
			for y in range(cy - radius - 1, cy + radius + 1):
				d = (x - cx)**2 + (y - cy)**2 
				if d < radius2:
					self.set_terrain(x, y, terrain)


	def place_random_forest(self, params):
		"""Places circular forest at random position.

		Returns nothing.
		"""

		forest = self.terraininfo.get_type("forest").tid

		radius = random.randint(*params.forests_radius)
		x = random.randint(radius + 1, self.w - radius - 1)
		y = random.randint(radius + 1, self.h - radius - 1)

		self.draw_terrain_circle(x, y, radius, forest)

	def create_river(self, params):
		width = random.randint(*params.river_width)
		w = self.w
		h = self.h

		wr = (w / 3, 2 * w / 3)
		hr = (h / 3, 2 * h / 3)

		ports = [
			((random.randint(*wr), 0), (random.randint(-wr[0], wr[0]), hr[1])),
			((random.randint(*hr), self.h - 1), (random.randint(-wr[0], wr[0]), -hr[0])),
			((0, random.randint(*hr)), (wr[0], random.randint(-hr[0], hr[0]))),
			((self.w - 1, random.randint(*hr)), (-wr[0], random.randint(-hr[0], hr[0])))]

		
		port_a = random.choice(ports)
		ports.remove(port_a)
		port_b = random.choice(ports)

		p_a, ofs = port_a
		p_b = p_a[0] + ofs[0], p_a[1] + ofs[1]
		p_d, ofs = port_b
		p_c = p_d[0] + ofs[0], p_d[1] + ofs[1]

		curve = bezier.Curve(p_a, p_b, p_c, p_d)

		grass = self.terraininfo.get_type("grass").tid
		water = self.terraininfo.get_type("water").tid

		for i in range(101):
			t = i / 100.
			x, y = curve.interpolate(t).tupi()
			self.draw_terrain_circle(x, y, width, water)

	def increment_cities(self, x, y):
		cr = self.terraininfo.city_radius
		xrange = range(max(0, x - cr), min(self.w, x + cr + 1))
		for ty in range(max(0, y - cr), min(self.h, y + cr + 1)):
			for tx in xrange:
				ov = self.cities[tx, ty]
				nv = max(cr - abs(tx - x) - abs(ty - y), 0)
				self.cities[tx, ty] = max(ov, nv)

	def set_terrain(self, x, y, v):
		"""Sets terrain type if inside boundaries, does nothing otherwise.
		"""
		if 0 <= x < self.w and 0 <= y < self.h:
			self.terrain[x, y] = v
			self.sprites[x, y] = v << 4
			self.dirty.append((x, y))
			for d in DIRECTIONS:
				nx, ny = d.calc(x, y)
				if 0 <= nx < self.w and 0 <= ny < self.h:
					self.dirty.append((nx, ny))
					if self.get_terrain(nx, ny) in self.neigh_map[v]:
						self.sprites[x, y] |= d.flag
						self.sprites[nx, ny] |= d.other
					else:
						self.sprites[nx, ny] &= ~d.other

	def set_terrain_owner(self, x, y, v, o):
		"""Sets terrain and owner."""
		self.set_terrain(x, y, v)
		if 0 <= x < self.w and 0 <= y < self.h:
			self.flags[x, y] &= ~OWNER_MASK
			self.flags[x, y] |= o

	def get_terrain(self, x, y, default=0):
		"""Get terrain, returns 0 if outside boundaries.
		"""
		if 0 <= x < self.w and 0 <= y < self.h:
			return self.terrain[x, y]
		else:
			return default

	def set_flags_rect(self, rect, v):
		flags = self.flags
		xrange = range(rect.left, rect.right)
		for y in range(rect.top, rect.bottom):
			for x in xrange:
				self.flags[x, y] |= v

	def set_owner(self, rect, owner):
		"""Takes care not to transfer land."""
		flags = self.flags
		owner = owner & OWNER_MASK
		xrange = range(rect.left, rect.right)
		for y in range(rect.top, rect.bottom):
			for x in xrange:
				if flags[x, y] & OWNER_MASK == 0:
					flags[x, y] |= owner
				self.dirty.append((x, y))

	def get_owner(self, x, y):
		return self.flags[x, y] & OWNER_MASK

	def get_rect_info(self, rect):
		owned = 0
		tobuy = 0
		nogood = 0
		city_mod = self.terraininfo.city_modifier_base
		type_map = dict.fromkeys(self.types.keys(), 0)

		flags = self.flags
		terrain = self.terrain
		xrange = range(rect.left, rect.right)

		for y in range(rect.top, rect.bottom):
			for x in xrange:
				owner = flags[x, y] & OWNER_MASK
				type_map[terrain[x, y]] += 1
				if owner == 1:
					type_map[terrain[x, y]] += 1
					owned += 1
				elif owner == 0:
					type_map[terrain[x, y]] += 1
					tobuy += city_mod ** self.cities[x, y]
				else:
					nogood += 1

		tobuy *= self.terraininfo.land_price
	
		return (rect, type_map, owned, tobuy), tobuy, nogood
	
	def get_project(self, target_tid, rinfo):
		rect, type_map, owned, tobuy = rinfo
		target = self.types[target_tid]
		area = sum(type_map.values())

		cost = tobuy

		ti = self.terraininfo
		target = ti.types[target_tid]
		natural = ti.get_type(target.natural)

		for tid, count in type_map.items():
			if tid == target_tid:
				continue

			cost += count * target.base_cost
			if tid != natural.tid:
				# Conversion cost.
				cost += count * (natural.base_cost + ti.types[tid].base_cost)

		return cost

if __name__ == '__main__':
	import pygame
	from pygame.locals import *

	pygame.init()

	game = Map(128, 128)

	game.generate_terrain()
	game.load_sprites()

	x = 0
	y = 0
	disp = pygame.display.set_mode((1024, 768), DOUBLEBUF)
	surf = pygame.Surface(game.get_prerender_size())
	game.prerender(surf)

	SPEED = 5

	keys = {}
	panning = 0
	running = 1
	clock = pygame.time.Clock()
	while running:
		for ev in pygame.event.get([KEYUP, KEYDOWN]):
			if ev.key == K_ESCAPE:
				running = 0

			keys[ev.key] = ev.type == KEYDOWN

		if panning:
			for ev in pygame.event.get(MOUSEMOTION):
				x += ev.rel[0]
				y += ev.rel[1]


		for ev in pygame.event.get([MOUSEBUTTONDOWN, MOUSEBUTTONUP]):
			if ev.button == 3:
				panning = ev.type == MOUSEBUTTONDOWN


		# Prevent event queue from clogging.
		pygame.event.clear()

		if keys.get(K_UP, 0):
			y += SPEED
		if keys.get(K_DOWN, 0):
			y -= SPEED
		if keys.get(K_LEFT, 0):
			x += SPEED
		if keys.get(K_RIGHT, 0):
			x -= SPEED

		disp.fill(0)
		disp.blit(surf, (x, y))
		pygame.display.flip()

